﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using Infragistics.Olap.FlatData;
using Infragistics.Win.UltraWinPivotGrid.UIElements;
using Infragistics.Win.UltraWinScrollBar;

namespace WPG_DataSetToFDS
{
    public partial class Form1 : Form
    {
        #region Fields
        const int rowHeight = 29;

        int originalWidth;
        int originalHeight;

        int lastColumnIndex = 0;
        int lastRowIndex = 0;

        bool firstPage = true;
        bool hasMoreColumns = false;
        bool hasMoreRows = false;

        float lastDrawnLocationX = 0;
        float lastDrawnLocationY = 0;

        Bitmap wholePivotImage = null;

        Dictionary<int, int> columnRectRightLocations = null;
        #endregion

        #region Pivot Setup
        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            DataTable table = GenerateData();
            IEnumerable<SportsData> data = CreateIEnumerable(table);
            FlatDataSource ds = InitializePivotData(data);

            var parameters = new CubeGenerationParameters();
            ds.GenerateCube(parameters);
            ds.InitializeAsync(ultraPivotGrid1);

            ultraPivotGrid1.DataSource = ds;
            olapDataSelector1.DataSource = ds;
            
            this.ultraPrintPreviewDialog1.Document = this.ultraPrintDocument1;
        }

        private DataTable GenerateData()
        {
            DataTable data = new DataTable();

            data.Columns.Add("Sport", typeof(string));
            data.Columns.Add("Nation", typeof(string));
            data.Columns.Add("City", typeof(string));
            data.Columns.Add("Budget", typeof(double));
            data.Columns.Add("Players", typeof(int));

            data.Rows.Add(new object[] { "Football", "England", "London", 31245, 5461});
            data.Rows.Add(new object[] { "Football", "England", "Manchester", 32514, 2134 });
            data.Rows.Add(new object[] { "Snooker", "England", "London", 12353, 5461 });
            data.Rows.Add(new object[] { "Tennis", "England", "London", 31245, 13532 });
            data.Rows.Add(new object[] { "Football", "France", "Paris", 5462, 12451 });
            data.Rows.Add(new object[] { "Tennis", "France", "Paris", 3151, 1245 });
            data.Rows.Add(new object[] { "Football", "Italy", "Milan", 6793, 5642 });
            data.Rows.Add(new object[] { "Volleyball", "Italy", "Rome", 31537, 3215 });
            data.Rows.Add(new object[] { "Basketball", "United States", "Miami", 3156, 16643 });
            data.Rows.Add(new object[] { "Basketball", "United States", "Los Angelees", 121467, 5461 });
            data.Rows.Add(new object[] { "Gymnastics", "China", "Beijing", 4123, 4789 });
            data.Rows.Add(new object[] { "Darts", "England", "London", 3129, 4314 });
            data.Rows.Add(new object[] { "Volleyball", "Bulgaria", "Sofia", 9690, 5431 });
            data.Rows.Add(new object[] { "Karate", "Japan", "Tokyo", 131254, 4231 });
            data.Rows.Add(new object[] { "Hockey", "Canada", "Toronto", 4325, 3123 });
            data.Rows.Add(new object[] { "Football", "Spain", "Madrid", 31551, 31235 });
            data.Rows.Add(new object[] { "Basketball", "Russia", "Moscow", 97473, 5461 });
            return data;
           
        }

        private FlatDataSource InitializePivotData(IEnumerable<SportsData> data)
        {
            IEnumerable<SportsData> itemSource = data;
            var settings = new FlatDataSourceInitialSettings();
            var ds = new FlatDataSource(itemSource, typeof(SportsData), settings);

            ds.InitializeHierarchyDescriptor += ds_InitializeHierarchyDescriptor;
            ds.InitializeMeasureDescriptor += ds_InitializeMeasureDescriptor;
            ds.AggregateMeasure += ds_AggregateMeasure;

            settings.Rows = "[SportsData].[Nation]";
            settings.Columns = "[SportsData].[Sport]";
            settings.Measures = "[Measures].[Budget]";

            return ds;
        }

        private IEnumerable<SportsData> CreateIEnumerable(DataTable table)
        {
            var listOfSportsData = table.Rows.Cast<DataRow>().Select(r => new SportsData(r.Field<string>("Sport"), 
                r.Field<string>("City"), r.Field<string>("Nation"), r.Field<double>("Budget"), 
                r.Field<int>("Players"))).ToList();
            return listOfSportsData;
        }

        private void ds_InitializeHierarchyDescriptor(object sender, InitializeHierarchyDescriptorEventArgs e)
        {
            LevelDescriptor level = e.HierarchyDescriptor.LevelDescriptors.ToList().FirstOrDefault(lvl => lvl.Name == "all");

            if (level != null)
            {
                switch (e.HierarchyDescriptor.PropertyDescriptor.Name)
                {
                    case "Nation":
                        level.MemberProvider = item => "All Nations";
                        break;
                    case "City":
                        level.MemberProvider = item => "All Cities";
                        break; 
                    case "Sport":
                        level.MemberProvider = item => "All Sports";
                        break; 
                }
            }

        }

        private void ds_InitializeMeasureDescriptor(object sender, InitializeMeasureDescriptorEventArgs e)
        {
            if ((e.MeasureDescriptor.Name == "Budget"))
            {
                e.MeasureDescriptor.DisplayFormat = "$#";
                e.MeasureDescriptor.Aggregation = MeasureAggregation.Sum;
            }
        }

        private void ds_AggregateMeasure(object sender, AggregateMeasureEventArgs e)
        {
            if (e.MeasureDescriptor.Name == "Budget")
            {
                double total = 0f;
                double count = 0f;
                foreach (object o in e.Items)
                {
                    dynamic item = o as SportsData;
                    if (item != null)
                    {
                        total += item.Budget;
                    }
                    count += 1f;
                }
                e.Value = total / count;
                e.Handled = true;
            }

        }
        #endregion

        #region Pivot Print
        private void button1_Click(object sender, EventArgs e)
        {
            ultraPrintPreviewDialog1.ShowDialog(this);
        }

        private void ultraPrintDocument1_BeginPrint(object sender, System.Drawing.Printing.PrintEventArgs e)
        {
            //this.ultraPrintDocument1.DefaultPageSettings.Landscape = true;

            this.ultraPrintDocument1.DefaultPageSettings.Margins.Left = 50;
            this.ultraPrintDocument1.DefaultPageSettings.Margins.Right = 50;
            this.ultraPrintDocument1.DefaultPageSettings.Margins.Top = 50;
            this.ultraPrintDocument1.DefaultPageSettings.Margins.Bottom = 50;

            this.ultraPrintDocument1.Header.TextCenter = "Pivot Grid Print Preview";
        }

        private void ultraPrintDocument1_PrintPage(object sender, System.Drawing.Printing.PrintPageEventArgs e)
        {
            if (firstPage)
            {
                // Get the image
                columnRectRightLocations = new Dictionary<int, int>();
                ExpandPivotIfNeeded();
                wholePivotImage = GetImageFromControl(ultraPivotGrid1);
                RestorePivot();
            }

            int minRowIndex = lastRowIndex;

            hasMoreColumns = false;
            hasMoreRows = false;

            lastDrawnLocationX = e.MarginBounds.Left;
            lastDrawnLocationY = e.MarginBounds.Top;

            var availableArea = new Rectangle(e.MarginBounds.Left, e.MarginBounds.Top,
                 e.MarginBounds.Right - e.MarginBounds.Left, e.MarginBounds.Bottom -e.MarginBounds.Top);

            DrawDropAreas(e.Graphics, availableArea);
            DrawHeaders(e.Graphics, availableArea);
            DrawNextCellRegion(e.Graphics, availableArea);

            if(hasMoreColumns)
            {
                lastRowIndex = minRowIndex;
            }
            else if(hasMoreRows)
            {
                lastColumnIndex = 0;
            }

            e.HasMorePages = hasMoreColumns || hasMoreRows;
        }

        private void DrawNextCellRegion(Graphics graphics, Rectangle availableArea)
        {
            var remainingArea = new RectangleF(lastDrawnLocationX, lastDrawnLocationY,
                    availableArea.Right - lastDrawnLocationX, availableArea.Bottom - lastDrawnLocationY);
            
            int drawnColumnsWidth;

            if (lastColumnIndex == 0)
            {
                drawnColumnsWidth = 0;
            }
            else
            {
                drawnColumnsWidth = columnRectRightLocations[lastColumnIndex -1];
            }

            for (int i = lastColumnIndex; i < ultraPivotGrid1.Columns.Count; i++)
            {
                lastColumnIndex = i;
                if (columnRectRightLocations[i] - drawnColumnsWidth > remainingArea.Width)
                {
                    hasMoreColumns = true;
                    break;
                }
            }

            var rowsToPrint = ultraPivotGrid1.Rows.Count - lastRowIndex;
            var drawnRowsHeight = lastRowIndex * rowHeight;

            if (rowsToPrint > 0)
            {
                hasMoreRows = remainingArea.Height < rowsToPrint * rowHeight;
            }
            else
            {
                Debug.WriteLine("No rows to print!");
                lastRowIndex = 0;
                hasMoreRows = false;
            }
            
            lastRowIndex = hasMoreRows ? lastRowIndex + (int)(remainingArea.Height / rowHeight) : ultraPivotGrid1.Rows.Count - 1;

            var cellArea = ultraPivotGrid1.UIElement.GetDescendant(typeof(CellAreaUIElement));

            if(cellArea == null)
            {
                Debug.WriteLine("Couldn't find the cell area element!");
                return;
            }

            var nextCellArea = new RectangleF(cellArea.Rect.Location.X + drawnColumnsWidth,
                                                 cellArea.Rect.Location.Y + drawnRowsHeight, 
                                                 remainingArea.Size.Width,
                                                 remainingArea.Size.Height);

            graphics.DrawImage(wholePivotImage, remainingArea,
               nextCellArea, GraphicsUnit.Pixel);

        }

        private void DrawHeaders(Graphics graphics, Rectangle availableArea)
        {
            DrawHeaderArea(graphics, availableArea, typeof(ColumnHeaderAreaUIElement), XYMove.Reset, XYMove.Yes);
            DrawHeaderArea(graphics, availableArea, typeof(RowHeaderAreaUIElement), XYMove.Yes, XYMove.No);
        }

        private void DrawDropAreas(Graphics graphics, Rectangle availableArea)
        {
            DrawDropArea(graphics, availableArea, typeof(FilterDropAreaUIElement), XYMove.No, XYMove.Yes);
            DrawDropArea(graphics, availableArea, typeof(MeasureDropAreaUIElement), XYMove.Yes, XYMove.No);
            DrawDropArea(graphics, availableArea, typeof(ColumnDropAreaUIElement), XYMove.Reset, XYMove.Yes);
            DrawDropArea(graphics, availableArea, typeof(RowDropAreaUIElement), XYMove.Yes, XYMove.No);
        }

        private void DrawHeaderArea(Graphics graphics, Rectangle availableArea, Type elementType, XYMove moveX, XYMove moveY)
        {
            var uiElement = ultraPivotGrid1.UIElement.GetDescendant(elementType);

            if (uiElement == null)
            {
                Debug.WriteLine("Failed to find element: " + elementType.Name);
                return;
            }
            var expandHeight = elementType == typeof(RowHeaderAreaUIElement) &&
                    uiElement.Rect.Bottom < uiElement.Parent.Rect.Bottom;
            
            var destinationRect = CalculateDrawingRect(uiElement.Rect, availableArea,expandHeight);

            RectangleF sourceRect;

            if (elementType == typeof(ColumnHeaderAreaUIElement))
            {
                var offset = lastColumnIndex == 0 ? 0 : columnRectRightLocations[lastColumnIndex - 1];
                sourceRect = new RectangleF(uiElement.Rect.Location.X + offset, 
                    uiElement.Rect.Location.Y, destinationRect.Size.Width, destinationRect.Size.Height);
            }
            else
            {
                sourceRect = new RectangleF(uiElement.Rect.Location.X, uiElement.Rect.Location.Y +
                        (lastRowIndex * rowHeight), destinationRect.Size.Width, destinationRect.Size.Height);
            }


            graphics.DrawImage(wholePivotImage, destinationRect,
                sourceRect, GraphicsUnit.Pixel);

            if (moveX == XYMove.Yes)
                lastDrawnLocationX += destinationRect.Width;
            else if (moveX == XYMove.Reset)
                lastDrawnLocationX = availableArea.Left;

            if (moveY == XYMove.Yes)
                lastDrawnLocationY += destinationRect.Height;
            else if (moveY == XYMove.Reset)
                lastDrawnLocationY += availableArea.Top;
        }

        private void DrawDropArea(Graphics graphics,Rectangle availableArea , Type elementType, XYMove moveX, XYMove moveY)
        {
            var uiElement = ultraPivotGrid1.UIElement.GetDescendant(elementType);

            if (uiElement == null)
            {
                Debug.WriteLine("Failed to find element: " + elementType.Name);
                return;
            }

            var destinationRect = CalculateDrawingRect(uiElement.Rect, availableArea, false);
            var sourceRect = new RectangleF(uiElement.Rect.Location, destinationRect.Size);
            

            graphics.DrawImage(wholePivotImage, destinationRect,
                sourceRect, GraphicsUnit.Pixel);

            if (moveX == XYMove.Yes)
                lastDrawnLocationX += destinationRect.Width;
            else if (moveX == XYMove.Reset)
                lastDrawnLocationX = availableArea.Left;

            if (moveY == XYMove.Yes)
                lastDrawnLocationY += destinationRect.Height;
            else if (moveY == XYMove.Reset)
                lastDrawnLocationY += availableArea.Top;
        }

        private RectangleF CalculateDrawingRect(Rectangle uiElementRectangle, Rectangle availableArea, bool expandHeight)
        {
            float width = uiElementRectangle.Width + lastDrawnLocationX > availableArea.Right ?
                availableArea.Right - lastDrawnLocationX :
                uiElementRectangle.Width;

            float height;

            if(expandHeight)
            {
                height = availableArea.Height;
            }
            else
            {
                height = uiElementRectangle.Height + lastDrawnLocationY > availableArea.Bottom ?
                availableArea.Bottom - lastDrawnLocationY :
                uiElementRectangle.Height;
            }

            return new RectangleF(lastDrawnLocationX, lastDrawnLocationY, width, height);
        }

        private void ExpandPivotIfNeeded()
        {
            bool needsUdating = false;

            ExpandWidthIfNeeded(needsUdating);
            ExpandHeightIfNeeded(needsUdating);

            if (needsUdating)
            {
                ultraPivotGrid1.Refresh();
            }
        }
        private void ExpandHeightIfNeeded(bool needsUpdating)
        {
            int newHeight = 0;

            newHeight += ultraPivotGrid1.Rows.Count * rowHeight;

            var columnHeaderArea = ultraPivotGrid1.UIElement.GetDescendant(typeof(ColumnHeaderAreaUIElement));
            if (columnHeaderArea != null)
            {
                newHeight += columnHeaderArea.Rect.Height;
            }

            var columnDropArea = ultraPivotGrid1.UIElement.GetDescendant(typeof(ColumnDropAreaUIElement));
            if (columnDropArea != null)
            {
                newHeight += columnDropArea.Rect.Height;
            }

            var filterDropArea = ultraPivotGrid1.UIElement.GetDescendant(typeof(FilterDropAreaUIElement));
            if (filterDropArea != null)
            {
                newHeight += filterDropArea.Rect.Height;
            }

            var scrollBar = ultraPivotGrid1.UIElement.
                GetDescendant(typeof(ScrollBarUIElement), Orientation.Horizontal);

            if (scrollBar != null)
            {
                newHeight += scrollBar.Rect.Height;
            }

            if (newHeight > ultraPivotGrid1.Height)
            {
                originalHeight = ultraPivotGrid1.Height;
                ultraPivotGrid1.Height = newHeight;
                needsUpdating = true;
            }

        }
        private void ExpandWidthIfNeeded(bool needsUdating)
        {
            int newWidth = 0;

            foreach (var column in ultraPivotGrid1.Columns)
            {
                newWidth += column.Width;
                columnRectRightLocations.Add(column.Index, newWidth);
            }
            //the row header area is awlays drawn so we can use the UIElement
            var rowHeaderArea = ultraPivotGrid1.UIElement.GetDescendant(typeof(RowHeaderAreaUIElement));
            if (rowHeaderArea != null)
            {
                newWidth += rowHeaderArea.Rect.Width;
            }
            //the scrollbar is also always drawn so we can also use its UIElement
            var scrollBar = ultraPivotGrid1.UIElement.
                GetDescendant(typeof(ScrollBarUIElement), Orientation.Vertical);

            if (scrollBar != null)
            {
                newWidth += scrollBar.Rect.Width;
            }

            if (newWidth > ultraPivotGrid1.Width)
            {
                originalWidth = ultraPivotGrid1.Width;
                ultraPivotGrid1.Width = newWidth;
                needsUdating = true;
            }
        }

        private Bitmap GetImageFromControl(Control control)
        {
            Bitmap bmp = new Bitmap(control.Width,control.Height);

            control.DrawToBitmap(bmp, new Rectangle (0,0 ,control.Width, control.Height));

            return bmp;
        }

        private void RestorePivot()
        {
            if (originalWidth != 0)
            {
                ultraPivotGrid1.Width = originalWidth;
                originalWidth = 0;
            }

            if (originalHeight != 0)
            {
                ultraPivotGrid1.Height = originalHeight;
                originalHeight = 0;
            }
        }

        private void ultraPrintDocument1_EndPrint(object sender, System.Drawing.Printing.PrintEventArgs e)
        {
            //restore settings
            firstPage = true;
            lastDrawnLocationX = 0;
            lastDrawnLocationY = 0;
            lastColumnIndex = 0;
            lastRowIndex = 0;
        }
#endregion

    }

    public enum XYMove
    {
        Yes, 
        No, 
        Reset
    }
}
